@@ -30,7 +30,6 @@ import stat
 import struct
 import sys
 import tempfile
-import warnings
 
 from dulwich.index import (
     Index,
@@ -64,9 +63,6 @@ from dulwich.tests import (
     TestCase,
     skipIf,
 )
-from dulwich.tests.utils import (
-    setup_warning_catcher,
-)
 
 
 def can_symlink():
@@ -75,10 +71,6 @@ def can_symlink():
         # Platforms other than Windows should allow symlinks without issues.
         return True
 
-    if not hasattr(os, "symlink"):
-        # Older Python versions do not have `os.symlink` on Windows.
-        return False
-
     test_source = tempfile.mkdtemp()
     test_target = test_source + "can_symlink"
     try:
@@ -90,7 +82,7 @@ def can_symlink():
 
 class IndexTestCase(TestCase):
 
-    datadir = os.path.join(os.path.dirname(__file__), "data/indexes")
+    datadir = os.path.join(os.path.dirname(__file__), "../../testdata/indexes")
 
     def get_simple_index(self, name):
         return Index(os.path.join(self.datadir, name))
@@ -109,26 +101,6 @@ class SimpleIndexTestCase(IndexTestCase):
             list(self.get_simple_index("index").iterobjects()),
         )
 
-    def test_iterblobs(self):
-        warnings.simplefilter("always", UserWarning)
-        self.addCleanup(warnings.resetwarnings)
-        warnings_list, restore_warnings = setup_warning_catcher()
-        self.addCleanup(restore_warnings)
-
-        self.assertEqual(
-            [(b"bla", b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391", 33188)],
-            list(self.get_simple_index("index").iterblobs()),
-        )
-
-        expected_warning = PendingDeprecationWarning("Use iterobjects() instead.")
-        for w in warnings_list:
-            if type(w) == type(expected_warning) and w.args == expected_warning.args:
-                break
-        else:
-            raise AssertionError(
-                "Expected warning %r not in %r" % (expected_warning, warnings_list)
-            )
-
     def test_getitem(self):
         self.assertEqual(
             (
